@@ -21,6 +21,7 @@
 #ifndef IEEE80211_H
 #define IEEE80211_H
 #include <linux/if_ether.h> /* ETH_ALEN */
+#include <linux/kernel.h>   /* ARRAY_SIZE */
 
 #define IEEE80211_DATA_LEN		2304
 /* Maximum size for the MA-UNITDATA primitive, 802.11 standard section
@@ -45,6 +46,42 @@ struct ieee80211_hdr {
 	u8 addr4[ETH_ALEN];
 } __attribute__ ((packed));
 
+struct ieee80211_hdr_3addr {
+	u16 frame_ctl;
+	u16 duration_id;
+	u8 addr1[ETH_ALEN];
+	u8 addr2[ETH_ALEN];
+	u8 addr3[ETH_ALEN];
+	u16 seq_ctl;
+} __attribute__ ((packed));
+
+static const char *eap_types[] = {
+	"EAP-Packet", "EAPOL-Start", "EAPOL-Logoff", "EAPOL-Key", 
+	"EAPOL-Encap-ASF-Alert", "Unknown"
+};
+
+static inline const char *eap_get_type(int type)
+{
+	if (type > ARRAY_SIZE(eap_types))
+		type = ARRAY_SIZE(eap_types) - 1;
+	return eap_types[type];
+}
+
+struct eapol {
+	u8 snap[6];
+	u16 ethertype;
+	u8 version;
+	u8 type;
+	u16 length;
+} __attribute__ ((packed));
+
+#define IEEE80211_3ADDR_LEN (24)
+#define IEEE80211_4ADDR_LEN (30)
+#define IEEE80211_FCS_LEN    (4)
+
+#define MIN_FRAG_THRESHOLD     256U
+#define	MAX_FRAG_THRESHOLD     2346U
+
 /* Frame control field constants */
 #define IEEE80211_FCTL_VERS		0x0002
 #define IEEE80211_FCTL_FTYPE		0x000c
@@ -140,6 +177,11 @@ do { if (ieee80211_debug_level & (level)) \
 #define IEEE80211_DL_SCAN          BIT(2)
 #define IEEE80211_DL_STATE         BIT(3)
 #define IEEE80211_DL_MGMT          BIT(4)
+#define IEEE80211_DL_FRAG          BIT(5)
+#define IEEE80211_DL_EAP           BIT(6)
+
+#define IEEE80211_DL_TX            BIT(8)
+#define IEEE80211_DL_RX            BIT(9)
 
 #define IEEE80211_ERROR(f, a...) printk(KERN_ERR "ieee80211: " f, ## a)
 #define IEEE80211_WARNING(f, a...) printk(KERN_WARNING "ieee80211: " f, ## a)
@@ -149,7 +191,10 @@ do { if (ieee80211_debug_level & (level)) \
 #define IEEE80211_DEBUG_SCAN(f, a...)   IEEE80211_DEBUG(IEEE80211_DL_SCAN, f, ## a)
 #define IEEE80211_DEBUG_STATE(f, a...)  IEEE80211_DEBUG(IEEE80211_DL_STATE, f, ## a)
 #define IEEE80211_DEBUG_MGMT(f, a...)  IEEE80211_DEBUG(IEEE80211_DL_MGMT, f, ## a)
-
+#define IEEE80211_DEBUG_FRAG(f, a...)  IEEE80211_DEBUG(IEEE80211_DL_FRAG, f, ## a)
+#define IEEE80211_DEBUG_EAP(f, a...)  IEEE80211_DEBUG(IEEE80211_DL_EAP, f, ## a)
+#define IEEE80211_DEBUG_TX(f, a...)  IEEE80211_DEBUG(IEEE80211_DL_TX, f, ## a)
+#define IEEE80211_DEBUG_RX(f, a...)  IEEE80211_DEBUG(IEEE80211_DL_RX, f, ## a)
 #include <linux/netdevice.h>
 #include <linux/wireless.h>
 #include <linux/if_arp.h> /* ARPHRD_ETHER */
@@ -160,11 +205,6 @@ do { if (ieee80211_debug_level & (level)) \
 #include <net/iw_handler.h>	// new driver API
 
 #define BIT(x) (1 << (x))
-
-#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
-#define MACSTR "%02x:%02x:%02x:%02x:%02x:%02x"
-
-
 #ifndef ETH_P_PAE
 #define ETH_P_PAE 0x888E /* Port Access Entity (IEEE 802.1X) */
 #endif /* ETH_P_PAE */
@@ -264,18 +304,80 @@ struct ieee80211_snap_hdr {
 #define IEEE80211_STATMASK_WEMASK 0x7
 
 
+#define IEEE80211_CCK_MODULATION    BIT(0)
+#define IEEE80211_OFDM_MODULATION   BIT(1)
+
+#define IEEE80211_24GHZ_BAND     BIT(0)
+#define IEEE80211_52GHZ_BAND     BIT(1)
+
+#define IEEE80211_CCK_RATE_1MB		        0x02	
+#define IEEE80211_CCK_RATE_2MB		        0x04	
+#define IEEE80211_CCK_RATE_5MB		        0x0B
+#define IEEE80211_CCK_RATE_11MB		        0x16
+#define IEEE80211_OFDM_RATE_6MB		        0x0C	
+#define IEEE80211_OFDM_RATE_9MB		        0x12
+#define IEEE80211_OFDM_RATE_12MB		0x18	
+#define IEEE80211_OFDM_RATE_18MB		0x24
+#define IEEE80211_OFDM_RATE_24MB		0x30
+#define IEEE80211_OFDM_RATE_36MB		0x48
+#define IEEE80211_OFDM_RATE_48MB		0x60
+#define IEEE80211_OFDM_RATE_54MB		0x6C
+#define IEEE80211_BASIC_RATE_MASK		0x80
+
+#define IEEE80211_CCK_RATE_1MB_MASK		BIT(0)	
+#define IEEE80211_CCK_RATE_2MB_MASK		BIT(1)
+#define IEEE80211_CCK_RATE_5MB_MASK		BIT(2)
+#define IEEE80211_CCK_RATE_11MB_MASK		BIT(3)
+#define IEEE80211_OFDM_RATE_6MB_MASK		BIT(4)
+#define IEEE80211_OFDM_RATE_9MB_MASK		BIT(5)
+#define IEEE80211_OFDM_RATE_12MB_MASK		BIT(6)
+#define IEEE80211_OFDM_RATE_18MB_MASK		BIT(7)
+#define IEEE80211_OFDM_RATE_24MB_MASK		BIT(8)	
+#define IEEE80211_OFDM_RATE_36MB_MASK		BIT(9)
+#define IEEE80211_OFDM_RATE_48MB_MASK		BIT(10)
+#define IEEE80211_OFDM_RATE_54MB_MASK		BIT(11)
+
+#define IEEE80211_CCK_RATES_MASK	        0x0000000F
+#define IEEE80211_CCK_BASIC_RATES_MASK	(IEEE80211_CCK_RATE_1MB_MASK | \
+	IEEE80211_CCK_RATE_2MB_MASK)
+#define IEEE80211_CCK_DEFAULT_RATES_MASK	(IEEE80211_CCK_BASIC_RATES_MASK | \
+        IEEE80211_CCK_RATE_5MB_MASK | \
+        IEEE80211_CCK_RATE_11MB_MASK)
+
+#define IEEE80211_OFDM_RATES_MASK		0x00000FF0
+#define IEEE80211_OFDM_BASIC_RATES_MASK	(IEEE80211_OFDM_RATE_6MB_MASK | \
+	IEEE80211_OFDM_RATE_12MB_MASK | \
+	IEEE80211_OFDM_RATE_24MB_MASK)
+#define IEEE80211_OFDM_DEFAULT_RATES_MASK	(IEEE80211_OFDM_BASIC_RATES_MASK | \
+	IEEE80211_OFDM_RATE_9MB_MASK  | \
+	IEEE80211_OFDM_RATE_18MB_MASK | \
+	IEEE80211_OFDM_RATE_36MB_MASK | \
+	IEEE80211_OFDM_RATE_48MB_MASK | \
+	IEEE80211_OFDM_RATE_54MB_MASK)
+#define IEEE80211_DEFAULT_RATES_MASK (IEEE80211_OFDM_DEFAULT_RATES_MASK | \
+                                IEEE80211_CCK_DEFAULT_RATES_MASK)
+
+#define IEEE80211_NUM_OFDM_RATES	    8
+#define IEEE80211_NUM_CCK_RATES	            4
+#define IEEE80211_OFDM_SHIFT_MASK_A         4
+
+
+
+
 /* NOTE: This data is for statistical purposes; not all hardware provides this
  *       information for frames received.  Not setting these will not cause
  *       any adverse affects. */
 struct ieee80211_rx_stats {
 	u32 mac_time;
-	u8 rssi;
+	s8 rssi;
 	u8 signal;
 	u8 noise;
 	u16 rate; /* in 100 kbps */
 	u8 received_channel;
 	u8 control;
 	u8 mask;
+	u8 freq;
+	u16 len;
 };
 
 /* IEEE 802.11 requires that STA supports concurrent reception of at least
@@ -341,7 +443,7 @@ struct ieee80211_device;
 #define WEP_KEY_LEN 13
 
 struct ieee80211_security {
-	u16 active_key:1, 
+	u16 active_key:2, 
             enabled:1, 
 	    auth_mode:2, 
             auth_algo:4, 
@@ -389,6 +491,7 @@ struct ieee80211_header_data {
 #define MFIE_TYPE_IBSS_SET   6
 #define MFIE_TYPE_CHALLENGE  16
 #define MFIE_TYPE_RSN	     48
+#define MFIE_TYPE_RATES_EX   50
 #define MFIE_TYPE_GENERIC    221
 
 struct ieee80211_info_element_hdr {
@@ -443,61 +546,58 @@ struct ieee80211_assoc_request_frame {
 	struct ieee80211_info_element info_element;
 } __attribute__ ((packed));
 
-struct ieee80211_helper_functions {
-	void (*set_security)(struct ieee80211_device *ieee, 
-			     struct ieee80211_security *sec);
-
-	/* these functions are defined in hardware model specific files
-	 * (hostap_{cs,plx,pci}.c */
-	int (*card_present)(struct ieee80211_device *ieee);
-	void (*cor_sreset)(struct ieee80211_device *ieee);
-	int (*dev_open)(struct ieee80211_device *ieee);
-	int (*dev_close)(struct ieee80211_device *ieee);
-	void (*genesis_reset)(struct ieee80211_device *ieee, int hcr);
-
-	
-	/* Turn on unencrypted packet filtering at the HW level */
-	void (*set_unencrypted_filter)(struct ieee80211_device *ieee, int flag);
-
-	/* the following functions are from hostap_hw.c, but they may have some
-	 * hardware model specific code */
-
-	int (*hw_enable)(struct net_device *dev, int initial);
-	int (*hw_config)(struct net_device *dev, int initial);
-	void (*hw_reset)(struct net_device *dev);
-	void (*hw_shutdown)(struct net_device *dev, int no_disable);
-	int (*reset_port)(struct net_device *dev);
-	int (*tx)(struct sk_buff *skb, struct net_device *dev);
-	void (*schedule_reset)(struct ieee80211_device *ieee);
-	int (*tx_80211)(struct sk_buff *skb, struct net_device *dev);
+struct ieee80211_txb {
+	u8 nr_frags;
+	u8 encrypted;
+	u16 reserved;
+	u16 frag_size;
+	u16 payload_size;
+	struct sk_buff *fragments[0];
 };
 
 
-
 /* SWEEP TABLE ENTRIES NUMBER*/
 #define MAX_SWEEP_TAB_ENTRIES		  42
 #define MAX_SWEEP_TAB_ENTRIES_PER_PACKET  7
+/* MAX_RATES_LENGTH needs to be 12.  The spec says 8, and many APs
+ * only use 8, and then use extended rates for the remaining supported
+ * rates.  Other APs, however, stick all of their supported rates on the
+ * main rates information element... */
 #define MAX_RATES_LENGTH                  ((u8)12)
-#define MAX_BEACON_COUNT                  128
+#define MAX_RATES_EX_LENGTH               ((u8)16)
+#define MAX_NETWORK_COUNT                  128
 
 #define CRC_LENGTH                 4U
 
 #define MAX_WPA_IE_LEN 64
 
-struct ieee80211_beacon {
+#define NETWORK_EMPTY_ESSID BIT(0)
+#define NETWORK_HAS_OFDM    BIT(1)
+#define NETWORK_HAS_CCK     BIT(2)
+
+struct ieee80211_network {
+	/* These entries are used to identify a unique network */
 	u8 bssid[ETH_ALEN];
-	u8 ssid_len;
-	struct ieee80211_rx_stats stats;
-	u16 capability;
 	u8 channel;
-	u8 rates[MAX_RATES_LENGTH];
-	u8 rates_len;
-	u32 last_scanned;
 	/* Ensure null-terminated for any debug msgs */
 	u8 ssid[IW_ESSID_MAX_SIZE + 1];
+	u8 ssid_len;
+
+	/* These are network statistics */
+	struct ieee80211_rx_stats stats;
+	u16 capability;
+	u8 rates[MAX_RATES_LENGTH];
+	u8 rates_len;
+	u8 rates_ex[MAX_RATES_EX_LENGTH];
+	u8 rates_ex_len;
+	unsigned long last_scanned;
+	u8 mode;
+	u8 flags;
+	u32 last_associate;
 	u32 time_stamp[2];
 	u16 beacon_interval;
 	u16 listen_interval;
+	u16 atim_window;
 #ifdef CONFIG_IEEE80211_WPA		
 	u8 wpa_ie[MAX_WPA_IE_LEN];
 	size_t wpa_ie_len;
@@ -517,8 +617,22 @@ enum ieee80211_state {
 	IEEE80211_SHUTDOWN
 };
 
-#define DEFAULT_MAX_SCAN_AGE 5
-#define DEFAULT_FTS 2342
+#define DEFAULT_MAX_SCAN_AGE (15 * HZ) 
+#define DEFAULT_FTS 2346
+#define MAC_FMT "%02x:%02x:%02x:%02x:%02x:%02x"
+#define MAC_ARG(x) ((u8*)(x))[0],((u8*)(x))[1],((u8*)(x))[2],((u8*)(x))[3],((u8*)(x))[4],((u8*)(x))[5]
+
+
+extern inline int is_multicast_ether_addr(const u8 *addr)
+{
+	return ((addr[0] != 0xff) && (0x01 & addr[0]));
+}
+
+extern inline int is_broadcast_ether_addr(const u8 *addr)
+{
+	return ((addr[0] == 0xff) && (addr[1] == 0xff) && (addr[2] == 0xff) &&   \
+		(addr[3] == 0xff) && (addr[4] == 0xff) && (addr[5] == 0xff));
+}
 
 struct ieee80211_device {
 	struct net_device *dev;
@@ -526,12 +640,11 @@ struct ieee80211_device {
 	/* Bookkeeping structures */
 	struct net_device_stats stats;
 	struct ieee80211_stats ieee_stats;
-	void *priv;
 
 	/* Probe / Beacon management */
-	struct list_head beacon_free_list;
-	struct list_head beacon_list;
-	struct ieee80211_beacon *beacons;
+	struct list_head network_free_list;
+	struct list_head network_list;
+	struct ieee80211_network *networks;
 	int scans;
 	int scan_age;
 
@@ -539,9 +652,6 @@ struct ieee80211_device {
 
 	spinlock_t lock;
 
-	int tx_payload_only; /* set to 1 if HW only expects the Tx SKB 
-			      * to contain the payload (vs. a fully configured
-			      * 802.11 header for a data frame) */
 	int tx_headroom; /* Set to size of any additional room needed at front
 			  * of allocated Tx SKBs */
 
@@ -554,8 +664,7 @@ struct ieee80211_device {
 	/* If the host performs {en,de}cryption, then set to 1 */
 	int host_encrypt;
 	int host_decrypt;
-
-	int ieee_802_1x; /* is IEEE 802.1X used */
+	int ieee802_1x; /* is IEEE 802.1X used */
 
 #ifdef CONFIG_IEEE80211_WPA
 	/* WPA data */
@@ -563,7 +672,6 @@ struct ieee80211_device {
 	int drop_unencrypted;
 	int tkip_countermeasures;
 	int privacy_invoked;
-	int pass_unencrypted_eapol;
 	size_t wpa_ie_len;
 	u8 *wpa_ie;
 #endif  /* CONFIG_IEEE80211_WPA */
@@ -584,14 +692,76 @@ struct ieee80211_device {
 	/* Association info */
 	u8 bssid[ETH_ALEN];
 
-	/* Callback vtable */
-	struct ieee80211_helper_functions *func;
-
 	enum ieee80211_state state;
 
+	int mode;       /* A, B, G */
+	int modulation; /* CCK, OFDM */
+	int freq_band;  /* 2.4Ghz, 5.2Ghz, Mixed */
+	int abg_ture;   /* ABG flag              */
 
+	/* Callback functions */
+	void (*set_security)(struct net_device *dev,
+			     struct ieee80211_security *sec);
+	int (*hard_start_xmit)(struct ieee80211_txb *txb, 
+			       struct net_device *dev);
+	int (*reset_port)(struct net_device *dev);
+
+	/* This must be the last item so that it points to the data 
+	 * allocated beyond this structure by alloc_ieee80211 */
+	u8 priv[0];
 };
 
+#define IEEE_A            BIT(0)
+#define IEEE_B            BIT(1)
+#define IEEE_G            BIT(2)
+#define IEEE_MODE_MASK    (IEEE_A|IEEE_B|IEEE_G)
+
+extern inline void *ieee80211_priv(struct net_device *dev)
+{
+	return ((struct ieee80211_device *)netdev_priv(dev))->priv;
+}
+
+extern inline int ieee80211_is_empty_essid(const char *essid, int essid_len)
+{
+	/* Single white space is for Linksys APs */
+	if (essid_len == 1 && essid[0] == ' ')
+		return 1;
+
+	/* Otherwise, if the entire essid is 0, we assume it is hidden */
+	while (essid_len) {
+		essid_len--;
+		if (essid[essid_len] != '\0')
+			return 0;
+	}
+
+	return 1;
+}
+
+extern inline int ieee80211_is_valid_mode(struct ieee80211_device *ieee, int mode)
+{
+	/*
+	 * It is possible for both access points and our device to support
+	 * combinations of modes, so as long as there is one valid combination
+	 * of ap/device supported modes, then return success
+	 *
+	 */
+	if ((mode & IEEE_A) && 
+	    (ieee->modulation & IEEE80211_OFDM_MODULATION) &&
+	    (ieee->freq_band & IEEE80211_52GHZ_BAND))
+		return 1;
+
+	if ((mode & IEEE_G) && 
+	    (ieee->modulation & IEEE80211_OFDM_MODULATION) &&
+	    (ieee->freq_band & IEEE80211_24GHZ_BAND))
+		return 1;
+
+	if ((mode & IEEE_B) && 
+	    (ieee->modulation & IEEE80211_CCK_MODULATION) &&
+	    (ieee->freq_band & IEEE80211_24GHZ_BAND))
+		return 1;
+
+	return 0;
+}
 
 extern inline int ieee80211_get_hdrlen(u16 fc)
 {
@@ -621,25 +791,16 @@ extern inline int ieee80211_get_hdrlen(u16 fc)
 
 
 /* ieee80211.c */
-extern struct ieee80211_device * ieee80211_alloc(struct net_device *dev,
-						 void *priv);
-extern void ieee80211_free(struct ieee80211_device *ieee);
+extern void free_ieee80211(struct net_device *dev);
+extern struct net_device *alloc_ieee80211(int sizeof_priv);
 
 extern int ieee80211_set_encryption(struct ieee80211_device *ieee);
 
 /* ieee80211_tx.c */
 
-struct ieee80211_txb {
-	u8 nr_frags;
-	u8 encrypted;
-	u16 reserved;
-	u16 frag_size;
-	u16 payload_size;
-	struct sk_buff *fragments[0];
-};
 
-extern struct ieee80211_txb *ieee80211_skb_to_txb(struct ieee80211_device *ieee, 
-						  struct sk_buff *skb);
+extern int ieee80211_xmit(struct sk_buff *skb,
+			  struct net_device *dev);
 extern void ieee80211_txb_free(struct ieee80211_txb *);
 
 
@@ -648,7 +809,6 @@ extern int ieee80211_rx(struct ieee80211_device *ieee, struct sk_buff *skb,
 			struct ieee80211_rx_stats *rx_stats);
 extern void ieee80211_rx_mgt(struct ieee80211_device *ieee,
 			     struct ieee80211_hdr *header,
-			     int frame_size,
 			     struct ieee80211_rx_stats *stats);
 
 /* iee80211_wx.c */
@@ -672,4 +832,33 @@ extern inline int ieee80211_get_scans(struct ieee80211_device *ieee)
 {
 	return ieee->scans;
 }
+
+static inline const char *escape_essid(const char *essid, u8 essid_len) {
+	static char escaped[IW_ESSID_MAX_SIZE * 2 + 1];
+	const char *s = essid;
+	char *d = escaped;
+
+	if (ieee80211_is_empty_essid(essid, essid_len)) {
+		memcpy(escaped, "<hidden>", sizeof("<hidden>"));
+		return escaped;
+	}
+		
+	essid_len = min(essid_len, (u8)IW_ESSID_MAX_SIZE);
+	while (essid_len--) {
+		if (*s == '\0') {
+			*d++ = '\\';
+			*d++ = '0';
+			s++;
+		} else {
+			*d++ = *s++;
+		}
+	}
+	*d = '\0';
+	return escaped;
+}
+
+#ifndef offset_in_page
+#define offset_in_page(p) ((unsigned long)(p) & ~PAGE_MASK)
+#endif
+
 #endif /* IEEE80211_H */
