@@ -273,7 +273,7 @@ static int ieee80211_is_eapol_frame(struct ieee80211_device *ieee,
 /* Called only as a tasklet (software IRQ), by ieee80211_rx */
 static inline int
 ieee80211_rx_frame_decrypt(struct ieee80211_device* ieee, struct sk_buff *skb,
-			struct ieee80211_crypt_data *crypt)
+			   struct ieee80211_crypt_data *crypt)
 {
 	struct ieee80211_hdr *hdr;
 	int res, hdrlen;
@@ -289,8 +289,8 @@ ieee80211_rx_frame_decrypt(struct ieee80211_device* ieee, struct sk_buff *skb,
 	    strcmp(crypt->ops->name, "TKIP") == 0) {
 		if (net_ratelimit()) {
 			printk(KERN_DEBUG "%s: TKIP countermeasures: dropped "
-			       "received packet from " MACSTR "\n",
-			       ieee->dev->name, MAC2STR(hdr->addr2));
+			       "received packet from " MAC_FMT "\n",
+			       ieee->dev->name, MAC_ARG(hdr->addr2));
 		}
 		return -1;
 	}
@@ -300,12 +300,13 @@ ieee80211_rx_frame_decrypt(struct ieee80211_device* ieee, struct sk_buff *skb,
 	res = crypt->ops->decrypt_mpdu(skb, hdrlen, crypt->priv);
 	atomic_dec(&crypt->refcnt);
 	if (res < 0) {
-		printk(KERN_DEBUG "%s: decryption failed (SA=" MACSTR
+		printk(KERN_DEBUG "%s: decryption failed (SA=" MAC_FMT
 		       ") res=%d\n",
-		       ieee->dev->name, MAC2STR(hdr->addr2), res);
+		       ieee->dev->name, MAC_ARG(hdr->addr2), res);
 		if (res == -2)
-			printk(KERN_DEBUG "%s: WEP decryption failed ICV mismatch\n",
-				ieee->dev->name);
+			printk(KERN_DEBUG "%s: WEP decryption failed ICV "
+			       "mismatch (key %d)\n",
+			       ieee->dev->name, skb->data[hdrlen + 3] >> 6);
 		ieee->ieee_stats.rx_discards_wep_undecryptable++;
 		return -1;
 	}
@@ -333,8 +334,8 @@ ieee80211_rx_frame_decrypt_msdu(struct ieee80211_device* ieee, struct sk_buff *s
 	atomic_dec(&crypt->refcnt);
 	if (res < 0) {
 		printk(KERN_DEBUG "%s: MSDU decryption/MIC verification failed"
-		       " (SA=" MACSTR " keyidx=%d)\n",
-		       ieee->dev->name, MAC2STR(hdr->addr2), keyidx);
+		       " (SA=" MAC_FMT " keyidx=%d)\n",
+		       ieee->dev->name, MAC_ARG(hdr->addr2), keyidx);
 		return -1;
 	}
 
@@ -450,8 +451,8 @@ int ieee80211_rx(struct ieee80211_device *ieee, struct sk_buff *skb,
 			 * frames silently instead of filling system log with
 			 * these reports. */
 			printk(KERN_DEBUG "%s: WEP decryption failed (not set)"
-			       " (SA=" MACSTR ")\n",
-			       ieee->dev->name, MAC2STR(hdr->addr2));
+			       " (SA=" MAC_FMT ")\n",
+			       ieee->dev->name, MAC_ARG(hdr->addr2));
 #endif
 			ieee->ieee_stats.rx_discards_wep_undecryptable++;
 			goto rx_dropped;
@@ -466,8 +467,8 @@ int ieee80211_rx(struct ieee80211_device *ieee, struct sk_buff *skb,
 		    (keyidx = hostap_rx_frame_decrypt(ieee, skb, crypt)) < 0)
 		{
 			printk(KERN_DEBUG "%s: failed to decrypt mgmt::auth "
-			       "from " MACSTR "\n", dev->name,
-			       MAC2STR(hdr->addr2));
+			       "from " MAC_FMT "\n", dev->name,
+			       MAC_ARG(hdr->addr2));
 			/* TODO: could inform hostapd about this so that it
 			 * could send auth failure report */
 			goto rx_dropped;
@@ -574,6 +575,8 @@ int ieee80211_rx(struct ieee80211_device *ieee, struct sk_buff *skb,
 	if ((frag != 0 || (fc & IEEE80211_FCTL_MOREFRAGS))) {
 		int flen;
 		struct sk_buff *frag_skb = ieee80211_frag_cache_get(ieee, hdr);
+		IEEE80211_DEBUG_FRAG("Rx Fragment received (%u)\n", frag);
+
 		if (!frag_skb) {
 			printk(KERN_DEBUG "%s: Rx cannot get skb from "
 			       "fragment cache (morefrag=%d seq=%u frag=%u)\n",
@@ -631,26 +634,38 @@ int ieee80211_rx(struct ieee80211_device *ieee, struct sk_buff *skb,
 
 	hdr = (struct ieee80211_hdr *) skb->data;
 	if (crypt && !(fc & IEEE80211_FCTL_WEP) && !ieee->open_wep) {
-		if (/*ieee->ieee_802_1x &&*/
+		if (/*ieee->ieee802_1x &&*/
 		    ieee80211_is_eapol_frame(ieee, skb)) {
 			/* pass unencrypted EAPOL frames even if encryption is
 			 * configured */
-			printk(KERN_DEBUG "%s: RX: IEEE 802.1X - passing "
-			       "unencrypted EAPOL frame\n", ieee->dev->name);
+			struct eapol *eap = (struct eapol *)(skb->data + 
+				24);
+			IEEE80211_DEBUG_EAP("RX: IEEE 802.1X EAPOL frame: %s\n",
+						eap_get_type(eap->type));
 		} else {
 			printk(KERN_DEBUG "%s: encryption configured, but RX "
-			       "frame not encrypted (SA=" MACSTR ")\n",
-			       ieee->dev->name, MAC2STR(hdr->addr2));
+			       "frame not encrypted (SA=" MAC_FMT ")\n",
+			       ieee->dev->name, MAC_ARG(hdr->addr2));
 			goto rx_dropped;
 		}
 	}
 
+#ifdef CONFIG_IEEE80211_DEBUG
+	if (crypt && !(fc & IEEE80211_FCTL_WEP) && 
+	    ieee80211_is_eapol_frame(ieee, skb)) {
+			struct eapol *eap = (struct eapol *)(skb->data + 
+				24);
+			IEEE80211_DEBUG_EAP("RX: IEEE 802.1X EAPOL frame: %s\n",
+						eap_get_type(eap->type));
+	}
+#endif
+
 	if (/*ieee->drop_unencrypted*/ crypt && !(fc & IEEE80211_FCTL_WEP) && !ieee->open_wep &&
 	    !ieee80211_is_eapol_frame(ieee, skb)) {
 		if (net_ratelimit()) {
 			printk(KERN_DEBUG "%s: dropped unencrypted RX data "
-			       "frame from " MACSTR " (drop_unencrypted=1)\n",
-			       dev->name, MAC2STR(hdr->addr2));
+			       "frame from " MAC_FMT " (drop_unencrypted=1)\n",
+			       dev->name, MAC_ARG(hdr->addr2));
 		}
 		goto rx_dropped;
 	}
@@ -664,7 +679,7 @@ int ieee80211_rx(struct ieee80211_device *ieee, struct sk_buff *skb,
 #ifdef NOT_YET
 	/* If IEEE 802.1X is used, check whether the port is authorized to send
 	 * the received frame. */
-	if (ieee->ieee_802_1x && ieee->iw_mode == IW_MODE_MASTER) {
+	if (ieee->ieee802_1x && ieee->iw_mode == IW_MODE_MASTER) {
 		if (ethertype == ETH_P_PAE) {
 			printk(KERN_DEBUG "%s: RX: IEEE 802.1X frame\n",
 			       dev->name);
@@ -779,214 +794,349 @@ int ieee80211_rx(struct ieee80211_device *ieee, struct sk_buff *skb,
 
 #define MGMT_FRAME_FIXED_PART_LENGTH		0x24
 
-static inline void ieee80211_beacon_add(
+static inline int ieee80211_is_ofdm_rate(u8 rate)
+{
+	switch (rate & ~IEEE80211_BASIC_RATE_MASK) {
+	case IEEE80211_OFDM_RATE_6MB:
+	case IEEE80211_OFDM_RATE_9MB:
+	case IEEE80211_OFDM_RATE_12MB:
+	case IEEE80211_OFDM_RATE_18MB:
+	case IEEE80211_OFDM_RATE_24MB:
+	case IEEE80211_OFDM_RATE_36MB:
+	case IEEE80211_OFDM_RATE_48MB:
+	case IEEE80211_OFDM_RATE_54MB:
+		return 1;
+	}
+        return 0;
+}
+
+
+static inline int ieee80211_network_init(
 	struct ieee80211_device *ieee,
-	struct ieee80211_probe_response *new_beacon,
-	int frame_size,
+	struct ieee80211_probe_response *beacon,
+	struct ieee80211_network *network,
 	struct ieee80211_rx_stats *stats)
 {
-	struct list_head *element;
-	struct ieee80211_beacon *beacon;
 	struct ieee80211_info_element *info_element;
-#ifdef CONFIG_IEEE80211_WPA		
- 	int left = frame_size - MGMT_FRAME_FIXED_PART_LENGTH;
-#endif /* CONFIG_IEEE80211_WPA */		
+ 	u16 left;
+	u8 i;
 
-	element = ieee->beacon_free_list.next;
-	
-	beacon = list_entry(element, struct ieee80211_beacon, list);
-	
 	/* Pull out fixed field data */
-	memcpy(beacon->bssid, new_beacon->header.addr3, ETH_ALEN);
-	beacon->capability = new_beacon->capability;
-	beacon->last_scanned = ieee->scans;
-	beacon->time_stamp[0] = new_beacon->time_stamp[0];
-	beacon->time_stamp[1] = new_beacon->time_stamp[1];
-	beacon->beacon_interval = new_beacon->beacon_interval;
-	beacon->listen_interval = 0x0A; /* Where to pull this? new_beacon->listen_interval;*/
+	memcpy(network->bssid, beacon->header.addr3, ETH_ALEN);
+	network->capability = beacon->capability;
+	network->last_scanned = jiffies;
+	network->time_stamp[0] = beacon->time_stamp[0];
+	network->time_stamp[1] = beacon->time_stamp[1];
+	network->beacon_interval = beacon->beacon_interval;
+	/* Where to pull this? beacon->listen_interval;*/
+	network->listen_interval = 0x0A; 
+	network->rates_len = network->rates_ex_len = 0;
+	network->last_associate = 0;
+	network->ssid_len = 0;
+	network->flags = 0;
+	network->atim_window = 0;
+
+	if (stats->freq == IEEE80211_52GHZ_BAND) {
+		/* for A band (No DS info) */
+		network->channel = stats->received_channel;
+	} else
+		network->flags |= NETWORK_HAS_CCK;
 
 #ifdef CONFIG_IEEE80211_WPA		
- 	beacon->wpa_ie_len = 0;
- 	beacon->rsn_ie_len = 0;
+ 	network->wpa_ie_len = 0;
+ 	network->rsn_ie_len = 0;
 #endif /* CONFIG_IEEE80211_WPA */		
 
-/*
-  if ((ieee->long_preamble == 0) &&
-  !(beacon->capability & WLAN_CAPABILITY_SHORT_PREAMBLE)) {
-  printk(KERN_INFO "%s: Beacon received from AP (" MACSTR ") "
-  "that has preable length conflict.\n",
-  ieee->ndev->name, MAC2STR(beacon->bssid));
-  }
-*/
-	/* First info element is the SSID */
-	info_element = &new_beacon->info_element;
-	if (info_element->id != MFIE_TYPE_SSID) {
-		IEEE80211_DEBUG(IPW_DL_INFO | IPW_DL_SCAN, 
-				"invalid MFIE type for SSID: %d\n",
-				info_element->id);
-		return;
-	}
-	beacon->ssid_len = min(info_element->len, (u8)IW_ESSID_MAX_SIZE);
-	memcpy(beacon->ssid, info_element->data, beacon->ssid_len);
-	if (beacon->ssid_len < IW_ESSID_MAX_SIZE)
-		memset(beacon->ssid + beacon->ssid_len, 0, 
-		       IW_ESSID_MAX_SIZE - beacon->ssid_len);
-
-#ifdef CONFIG_IEEE80211_WPA		
- 	left -= 2+info_element->len;
-#endif /* CONFIG_IEEE80211_WPA */		
-
-	/* Second info element is the RATES */
-	info_element = (struct ieee80211_info_element *)
-		&info_element->data[info_element->len];
-	if (info_element->id != MFIE_TYPE_RATES) {
-		IEEE80211_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
-				"invalid MFIE type for RATES: %d\n",
-				info_element->id);
-		return;
-	}
-	beacon->rates_len = min(info_element->len, MAX_RATES_LENGTH);
-	memcpy(beacon->rates, info_element->data, beacon->rates_len);
-
-#ifdef CONFIG_IEEE80211_WPA		
- 	left -= 2+info_element->len;
-#endif /* CONFIG_IEEE80211_WPA */		
-
-	/* Third info element is either FH or DS, based on BSS capability */
-	info_element = (struct ieee80211_info_element *)
-		&info_element->data[info_element->len];
-	if (beacon->capability & WLAN_CAPABILITY_BSS) {
-		/* BSS - Direct Sequence */
-		if ( stats->control != 0 ) {
-			if (info_element->id != MFIE_TYPE_DS_SET) {
-				IEEE80211_DEBUG(
-					IPW_DL_INFO | IPW_DL_SCAN,
-					": invalid MFIE type for DIRECT SEQUENCE: "
-					"%d\n",
-					info_element->id);
-				return;
+ 	info_element = &beacon->info_element;
+	left = stats->len - ((void *)info_element - (void *)beacon);
+	while (left >= sizeof(struct ieee80211_info_element_hdr)) {
+		if (sizeof(struct ieee80211_info_element_hdr) + info_element->len > left) {
+			IEEE80211_DEBUG_SCAN("SCAN: parse failed: info_element->len + 2 > left : info_element->len+2=%d left=%d.\n",
+					     info_element->len + sizeof(struct ieee80211_info_element),
+					     left);
+			return 1;
+               	}
+  		
+		switch (info_element->id) {
+		case MFIE_TYPE_SSID:
+			if (ieee80211_is_empty_essid(info_element->data,
+						     info_element->len)) {
+				network->flags |= NETWORK_EMPTY_ESSID;
+				break;
 			}
-		
-			beacon->channel = info_element->data[0];
-		} else {
-			/* for A band (No DS info) */
-			beacon->channel = stats->received_channel;
-		}
 
-#ifdef CONFIG_IEEE80211_WPA
-		while (left >= 2 && info_element->len < left){
-			info_element = (struct ieee80211_info_element *)
-				&info_element->data[info_element->len];
-			if (info_element->id == MFIE_TYPE_GENERIC &&
-			    info_element->len >= 4  &&
+			network->ssid_len = min(info_element->len, 
+						(u8)IW_ESSID_MAX_SIZE);
+			memcpy(network->ssid, info_element->data, network->ssid_len);
+        		if (network->ssid_len < IW_ESSID_MAX_SIZE)
+                		memset(network->ssid + network->ssid_len, 0,
+				       IW_ESSID_MAX_SIZE - network->ssid_len);
+			
+			IEEE80211_DEBUG_SCAN("MFIE_TYPE_SSID: '%s' len=%d.\n", 
+					     network->ssid, network->ssid_len);
+			break;
+
+		case MFIE_TYPE_RATES:
+			network->rates_len = min(info_element->len, MAX_RATES_LENGTH);
+			for (i = 0; i < network->rates_len; i++) {
+				network->rates[i] = info_element->data[i];
+				if (ieee80211_is_ofdm_rate(info_element->data[i])) {
+					network->flags |= NETWORK_HAS_OFDM;
+					if (info_element->data[i] & 
+					    IEEE80211_BASIC_RATE_MASK)
+						network->flags &= 
+							~NETWORK_HAS_CCK;
+				}
+			}
+			break;
+
+		case MFIE_TYPE_RATES_EX:
+			network->rates_ex_len = min(info_element->len, MAX_RATES_EX_LENGTH);
+			for (i = 0; i < network->rates_ex_len; i++) {
+				network->rates_ex[i] = info_element->data[i];
+				if (ieee80211_is_ofdm_rate(info_element->data[i])) {
+					network->flags |= NETWORK_HAS_OFDM;
+					if (info_element->data[i] & 
+					    IEEE80211_BASIC_RATE_MASK)
+						network->flags &= 
+							~NETWORK_HAS_CCK;
+				}
+			}
+			break;
+
+		case MFIE_TYPE_DS_SET:
+  			IEEE80211_DEBUG_SCAN("MFIE_TYPE_DS_SET: %d\n",
+					     info_element->data[0]);
+			if (stats->freq == IEEE80211_24GHZ_BAND)
+				network->channel = info_element->data[0];
+			break;
+
+	 	case MFIE_TYPE_FH_SET:
+  			IEEE80211_DEBUG_SCAN("MFIE_TYPE_FH_SET: ignored\n");
+			break;
+
+		case MFIE_TYPE_CF_SET:
+			IEEE80211_DEBUG_SCAN("MFIE_TYPE_CF_SET: ignored\n");
+			break;
+
+		case MFIE_TYPE_TIM:
+			IEEE80211_DEBUG_SCAN("MFIE_TYPE_TIM: ignored\n");
+			break;
+
+		case MFIE_TYPE_IBSS_SET:
+			IEEE80211_DEBUG_SCAN("MFIE_TYPE_IBSS_SET: ignored\n");
+			break;
+
+		case MFIE_TYPE_CHALLENGE:
+			IEEE80211_DEBUG_SCAN("MFIE_TYPE_CHALLENGE: ignored\n");
+			break;
+
+#ifdef CONFIG_IEEE80211_WPA		
+		case MFIE_TYPE_GENERIC:
+			IEEE80211_DEBUG_SCAN("MFIE_TYPE_GENERIC: %d bytes\n",
+					     info_element->len);
+			if (info_element->len >= 4  &&
 			    info_element->data[0] == 0x00 &&
 			    info_element->data[1] == 0x50 &&
 			    info_element->data[2] == 0xf2 &&
-			    info_element->data[3] == 0x01
-				){
-				beacon->wpa_ie_len = min(info_element->len + 2,MAX_WPA_IE_LEN);
-				memcpy(beacon->wpa_ie, info_element, beacon->wpa_ie_len);
-
-			} else if (info_element->id == MFIE_TYPE_RSN){
-				beacon->rsn_ie_len = min(info_element->len + 2,MAX_WPA_IE_LEN);
-				memcpy(beacon->rsn_ie, info_element, beacon->rsn_ie_len);
+			    info_element->data[3] == 0x01) {
+				network->wpa_ie_len = min(info_element->len + 2,
+							 MAX_WPA_IE_LEN);
+				memcpy(network->wpa_ie, info_element, 
+				       network->wpa_ie_len);
 			}
-			left -= (2+info_element->len);
-		}
-#endif /* CONFIG_IEEE80211_WPA */
-
-
-	} else if (beacon->capability & WLAN_CAPABILITY_IBSS) {
-		/* IBSS - Frequency Hop */
-#ifdef NOT_YET
-		if (info_element->id != MFIE_TYPE_FH_SET) {
-			IEEE80211_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
-					"invalid MFIE type for FREQ HOP: %d\n",
-					info_element->id);
-			return;
-		}
+			break;
+			
+		case MFIE_TYPE_RSN:
+			IEEE80211_DEBUG_SCAN("MFIE_TYPE_RSN: %d bytes\n",
+					     info_element->len);
+			network->rsn_ie_len = min(info_element->len + 2,
+						 MAX_WPA_IE_LEN);
+			memcpy(network->rsn_ie, info_element, 
+			       network->rsn_ie_len);
+			break;
 #endif
+
+		default:
+			IEEE80211_DEBUG_SCAN("unsupported IE %d\n", 
+					     info_element->id);
+                        break;
+  		}
+
+		left -= sizeof(struct ieee80211_info_element_hdr) + 
+			info_element->len;
+		info_element = (struct ieee80211_info_element *)
+                	&info_element->data[info_element->len];
+  	}
+	
+	network->mode = 0;
+	if (stats->freq == IEEE80211_52GHZ_BAND) {
+		network->mode = IEEE_A;
 	} else {
-		/* Invalid ESS/IBSS capability! */
-		IEEE80211_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
-				"invalid ESS/IBSS capability (both 0).\n");
-		return;
+		if (network->flags & NETWORK_HAS_OFDM)
+			network->mode |= IEEE_G;
+		if (network->flags & NETWORK_HAS_CCK)
+			network->mode |= IEEE_B;
 	}
 
-	/* For now, we ignore the optional CF and IBSS parameter sets */
+	if (network->mode == 0) {
+		IEEE80211_DEBUG_SCAN("Filtered out '%s (" MAC_FMT ")' "
+				     "network.\n",
+				     escape_essid(network->ssid, 
+						  network->ssid_len),
+				     MAC_ARG(network->bssid));
+		return 1;
+	}
+	
+	if (ieee80211_is_empty_essid(network->ssid, network->ssid_len))
+		network->flags |= NETWORK_EMPTY_ESSID;
 
-	list_del(element);
+	memcpy(&network->stats, stats, sizeof(network->stats));
 
-	IEEE80211_DEBUG_SCAN("adding '%s' to beacon list.\n", beacon->ssid);
-	memcpy(&beacon->stats, stats, sizeof(beacon->stats));
+	return 0;
+}
 
-	list_add_tail(&beacon->list, &ieee->beacon_list);
+static inline int is_same_network(struct ieee80211_network *src,
+				  struct ieee80211_network *dst)
+{
+	/* A network is only a duplicate if the channel, BSSID, and ESSID
+	 * all match.  We treat all <hidden> with the same BSSID and channel
+	 * as one network */
+	return ((src->ssid_len == dst->ssid_len) &&
+		(src->channel == dst->channel) &&
+		!memcmp(src->bssid, dst->bssid, ETH_ALEN) &&
+		!memcmp(src->ssid, dst->ssid, src->ssid_len));
+}
+
+static inline void update_network(struct ieee80211_network *dst,
+				  struct ieee80211_network *src)
+{
+	memcpy(&dst->stats, &src->stats, sizeof(struct ieee80211_rx_stats));
+	dst->capability = src->capability;
+	memcpy(dst->rates, src->rates, src->rates_len);
+	dst->rates_len = src->rates_len;
+	memcpy(dst->rates_ex, src->rates_ex, src->rates_ex_len);
+	dst->rates_ex_len = src->rates_ex_len;
+
+	dst->mode = src->mode;
+	dst->flags = src->flags;
+	dst->time_stamp[0] = src->time_stamp[0];
+	dst->time_stamp[1] = src->time_stamp[1];
+
+	dst->beacon_interval = src->beacon_interval;
+	dst->listen_interval = src->listen_interval;
+	dst->atim_window = src->atim_window;
+
+#ifdef CONFIG_IEEE80211_WPA		
+	memcpy(dst->wpa_ie, src->wpa_ie, src->wpa_ie_len);
+	dst->wpa_ie_len = src->wpa_ie_len;
+	memcpy(dst->rsn_ie, src->rsn_ie, src->rsn_ie_len);
+	dst->rsn_ie_len = src->rsn_ie_len;
+#endif /* CONFIG_IEEE80211_WPA */	
+
+	dst->last_scanned = jiffies;
+	/* dst->last_associate is not overwritten */
 }
 
 static inline void ieee80211_process_probe_response(
 	struct ieee80211_device *ieee,
-	struct ieee80211_probe_response *new_beacon,
-	int frame_size,
+	struct ieee80211_probe_response *beacon,
 	struct ieee80211_rx_stats *stats)
 {
-	struct ieee80211_info_element *ssid_info;
-	struct ieee80211_beacon *beacon;
-	u8 ssid_len;
+	struct ieee80211_network network;
+	struct ieee80211_network *target;
+	struct ieee80211_network *oldest = NULL;
+#ifdef CONFIG_IEEE80211_DEBUG
+	struct ieee80211_info_element *info_element = &beacon->info_element;
+#endif
 
 	IEEE80211_DEBUG_SCAN(
-		"\n"
-		"Time Stamp 1    : %08X\n"
-		"Time Stamp 2    : %08X\n"
-		"Beacon Interval : %04X\n"
-		"Capabilities    : %c%c%c%c-%c%c%c%c\n",
-		new_beacon->time_stamp[0],
+		"'%s' (" MAC_FMT "): %c%c%c%c-%c%c%c%c\n",
+		escape_essid(info_element->data, info_element->len),
+		MAC_ARG(beacon->header.addr3),
+		(beacon->capability & BIT(7)) ? '1' : '0',
+		(beacon->capability & BIT(6)) ? '1' : '0',
+		(beacon->capability & BIT(5)) ? '1' : '0',
+		(beacon->capability & BIT(4)) ? '1' : '0',
+		(beacon->capability & BIT(3)) ? '1' : '0',
+		(beacon->capability & BIT(2)) ? '1' : '0',
+		(beacon->capability & BIT(1)) ? '1' : '0',
+		(beacon->capability & BIT(0)) ? '1' : '0');
 
-		new_beacon->time_stamp[1],
-		new_beacon->beacon_interval,
-		(new_beacon->capability & BIT(7)) ? '1' : '0',
-		(new_beacon->capability & BIT(6)) ? '1' : '0',
-		(new_beacon->capability & BIT(5)) ? '1' : '0',
-		(new_beacon->capability & BIT(4)) ? '1' : '0',
-		(new_beacon->capability & BIT(3)) ? '1' : '0',
-		(new_beacon->capability & BIT(2)) ? '1' : '0',
-		(new_beacon->capability & BIT(1)) ? '1' : '0',
-		(new_beacon->capability & BIT(0)) ? '1' : '0');
-	
-	ssid_info = &new_beacon->info_element;
-	if (ssid_info->id != MFIE_TYPE_SSID) {
-		IEEE80211_DEBUG(IPW_DL_INFO | IPW_DL_SCAN,
-			  "invalid MFIE type for SSID: %d\n",
-			  ssid_info->id);
+	if (ieee80211_network_init(ieee, beacon, &network, stats)) {
+		IEEE80211_DEBUG_SCAN("Dropped '%s' (" MAC_FMT ") via %s.\n",
+				     escape_essid(info_element->data, 
+						  info_element->len),
+				     MAC_ARG(beacon->header.addr3),
+				     WLAN_FC_GET_STYPE(beacon->header.frame_ctl) ==
+				     IEEE80211_STYPE_PROBE_RESP ? 
+				     "PROBE RESPONSE" : "BEACON");
 		return;
 	}
 
-	/* Search for this entry in the list and nuke it if it is 
-	 * already there */
-	ssid_len = min(ssid_info->len, (u8)IW_ESSID_MAX_SIZE);
-	list_for_each_entry(beacon, &ieee->beacon_list, list) {
-		if (!memcmp(beacon->bssid, new_beacon->header.addr3, 
-			    ETH_ALEN) &&
-		    !memcmp(beacon->ssid, ssid_info->data,
-			    min(beacon->ssid_len, ssid_len))) {
-			list_del(&beacon->list);
-			list_add_tail(&beacon->list, &ieee->beacon_free_list);
+	/* The network parsed correctly -- so now we scan our known networks
+	 * to see if we can find it in our list.
+	 *  
+	 * NOTE:  This search is definitely not optimized.  Once its doing
+	 *        the "right thing" we'll optimize it for efficiency if 
+	 *        necessary */
+	
+	/* Search for this entry in the list and update it if it is 
+	 * already there. */
+	list_for_each_entry(target, &ieee->network_list, list) {
+		if (is_same_network(target, &network))
 			break;
-		}
+
+		if ((oldest == NULL) || 
+		    (target->last_scanned < oldest->last_scanned))
+			oldest = target;
 	}
 
-	if (list_empty(&ieee->beacon_free_list)) {
-		IEEE80211_DEBUG_INFO("TODO: Expire oldest beacon to add new.\n");
-		IEEE80211_WARNING("%s: Beacon free list is empty.\n",
-				  ieee->dev->name);
-	} else 
-		ieee80211_beacon_add(ieee, new_beacon, frame_size, stats);
+	/* If we didn't find a match, then get a new network slot to initialize
+	 * with this beacon's information */
+	if (&target->list == &ieee->network_list) {
+		if (list_empty(&ieee->network_free_list)) {
+			/* If there are no more slots, expire the oldest */
+			list_del(&oldest->list);
+			target = oldest;
+			IEEE80211_DEBUG_SCAN("Expired '%s' (" MAC_FMT ") from "
+					     "network list.\n", 
+					     escape_essid(target->ssid, 
+							  target->ssid_len),
+					     MAC_ARG(target->bssid));
+		} else {
+			/* Otherwise just pull from the free list */
+			target = list_entry(ieee->network_free_list.next, 
+					    struct ieee80211_network, list);
+			list_del(ieee->network_free_list.next);
+		}
 
-	IEEE80211_DEBUG_SCAN("exit\n");
+		
+#ifdef CONFIG_IEEE80211_DEBUG
+		IEEE80211_DEBUG_SCAN("Adding '%s' (" MAC_FMT ") via %s.\n",
+				     escape_essid(network.ssid, 
+						  network.ssid_len),
+				     MAC_ARG(network.bssid),
+				     WLAN_FC_GET_STYPE(beacon->header.frame_ctl) ==
+				     IEEE80211_STYPE_PROBE_RESP ? 
+				     "PROBE RESPONSE" : "BEACON");
+#endif
+		memcpy(target, &network, sizeof(*target));
+		list_add_tail(&target->list, &ieee->network_list);
+	} else {
+		IEEE80211_DEBUG_SCAN("Updating '%s' (" MAC_FMT ") via %s.\n",
+				     escape_essid(target->ssid, 
+						  target->ssid_len),
+				     MAC_ARG(target->bssid),
+				     WLAN_FC_GET_STYPE(beacon->header.frame_ctl) ==
+				     IEEE80211_STYPE_PROBE_RESP ? 
+				     "PROBE RESPONSE" : "BEACON");
+		update_network(target, &network);
+	}
 }
 	
 void ieee80211_rx_mgt(struct ieee80211_device *ieee,
 		      struct ieee80211_hdr *header,
-		      int frame_size,
 		      struct ieee80211_rx_stats *stats)
 {
 	switch (WLAN_FC_GET_STYPE(header->frame_ctl)) {
@@ -1005,7 +1155,7 @@ void ieee80211_rx_mgt(struct ieee80211_device *ieee,
 				     WLAN_FC_GET_STYPE(header->frame_ctl));
 		IEEE80211_DEBUG_SCAN("Probe response\n");
 		ieee80211_process_probe_response(
-			ieee, (struct ieee80211_probe_response *)header, frame_size, stats);
+			ieee, (struct ieee80211_probe_response *)header, stats);
 		break;
 
 	case IEEE80211_STYPE_BEACON:
@@ -1013,7 +1163,7 @@ void ieee80211_rx_mgt(struct ieee80211_device *ieee,
 				     WLAN_FC_GET_STYPE(header->frame_ctl));
 		IEEE80211_DEBUG_SCAN("Beacon\n");
 		ieee80211_process_probe_response(
-			ieee, (struct ieee80211_probe_response *)header, frame_size, stats);
+			ieee, (struct ieee80211_probe_response *)header, stats);
 		break;
 
 	default:
