@@ -42,18 +42,7 @@
 
 #include "ieee80211.h"
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
-#include <linux/sched.h>
-#define work_struct tq_struct
-#define schedule_work schedule_task
-#define INIT_WORK INIT_TQUEUE
-#define get_seconds() CURRENT_TIME
-#define queue_work(x,y) schedule_task(y)
-static inline struct workqueue_struct *create_workqueue(char *x) { return NULL; }
-#define destroy_workqueue(x) do {} while (0)
-#else
 #include <linux/workqueue.h>
-#endif
 
 #ifndef IRQ_NONE
 typedef void irqreturn_t;
@@ -62,14 +51,10 @@ typedef void irqreturn_t;
 #define IRQ_RETVAL(x)
 #endif
 
-#if ! (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,4) || \
-       (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,27) && \
-        LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)) )
-static inline void *netdev_priv(struct net_device *dev) { return dev->priv; }
-#endif
-
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,22)
-#define pci_name(dev)	((dev)->slot_name)
+#if WIRELESS_EXT < 17
+#define IW_QUAL_QUAL_INVALID  0x10
+#define IW_QUAL_LEVEL_INVALID 0x20
+#define IW_QUAL_NOISE_INVALID 0x40
 #endif
 
 #if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,6,5) )
@@ -89,7 +74,7 @@ struct ipw2100_rx_packet;
 
 #ifdef CONFIG_IPW_DEBUG
 enum { IPW_DEBUG_ENABLED = 1 };
-extern unsigned long ipw2100_debug_level;
+extern u32 ipw2100_debug_level;
 #define IPW_DEBUG(level, message...) \
 do { \
 	if (ipw2100_debug_level & (level)) { \
@@ -136,43 +121,51 @@ enum { IPW_DEBUG_ENABLED = 0 };
 #define IPW_DL_ERROR         BIT(0)
 #define IPW_DL_WARNING       BIT(1)
 #define IPW_DL_INFO          BIT(2)
-#define IPW_DL_WIRELESS_EXT  BIT(3)
-#define IPW_DL_ISR           BIT(4)
-#define IPW_DL_SCAN          BIT(5)
-#define IPW_DL_STATUS        BIT(6)
-#define IPW_DL_TRACE         BIT(7)
-#define IPW_DL_RX            BIT(8)
-#define IPW_DL_TX            BIT(9)
-#define IPW_DL_MANAGE        BIT(10)
-#define IPW_DL_WEP           BIT(11)
-#define IPW_DL_FW_COMMAND    BIT(12)
-#define IPW_DL_FRAG          BIT(13)
-#define IPW_DL_FW            BIT(14)
-#define IPW_DL_RF_KILL       BIT(15)
-#define IPW_DL_DROP          BIT(16)
-#define IPW_DL_IO            BIT(17)
-#define IPW_DL_IOCTL         BIT(18)
-#define IPW_DL_TEST          BIT(19)
+#define IPW_DL_WX            BIT(3)
+#define IPW_DL_HC            BIT(5)
+#define IPW_DL_STATE         BIT(6)
+
+#define IPW_DL_NOTIF         BIT(10)
+#define IPW_DL_SCAN          BIT(11)
+#define IPW_DL_ASSOC         BIT(12)
+#define IPW_DL_DROP          BIT(13)
+
+#define IPW_DL_IOCTL         BIT(14)
+#define IPW_DL_RF_KILL       BIT(17)
+
+
+#define IPW_DL_MANAGE        BIT(15)
+#define IPW_DL_FW            BIT(16)
+
+#define IPW_DL_FRAG          BIT(21)
+#define IPW_DL_WEP           BIT(22)
+#define IPW_DL_TX            BIT(23)
+#define IPW_DL_RX            BIT(24)
+#define IPW_DL_ISR           BIT(25)
+#define IPW_DL_IO            BIT(26)
+#define IPW_DL_TRACE         BIT(28)
 
 #define IPW_DEBUG_ERROR(f, a...) printk(KERN_ERR DRV_NAME ": " f, ## a)
 #define IPW_DEBUG_WARNING(f, a...) printk(KERN_WARNING DRV_NAME ": " f, ## a)
 #define IPW_DEBUG_INFO(f...)    IPW_DEBUG(IPW_DL_INFO, ## f)
-#define IPW_DEBUG_WX(f...)     IPW_DEBUG(IPW_DL_WIRELESS_EXT, ## f)
+#define IPW_DEBUG_WX(f...)     IPW_DEBUG(IPW_DL_WX, ## f)
 #define IPW_DEBUG_SCAN(f...)   IPW_DEBUG(IPW_DL_SCAN, ## f)
-#define IPW_DEBUG_STATUS(f...) IPW_DEBUG(IPW_DL_STATUS, ## f)
+#define IPW_DEBUG_NOTIF(f...) IPW_DEBUG(IPW_DL_NOTIF, ## f)
 #define IPW_DEBUG_TRACE(f...)  IPW_DEBUG(IPW_DL_TRACE, ## f)
 #define IPW_DEBUG_RX(f...)     IPW_DEBUG(IPW_DL_RX, ## f)
 #define IPW_DEBUG_TX(f...)     IPW_DEBUG(IPW_DL_TX, ## f)
 #define IPW_DEBUG_ISR(f...)    IPW_DEBUG(IPW_DL_ISR, ## f)
 #define IPW_DEBUG_MANAGEMENT(f...) IPW_DEBUG(IPW_DL_MANAGE, ## f)
 #define IPW_DEBUG_WEP(f...)    IPW_DEBUG(IPW_DL_WEP, ## f)
-#define IPW_DEBUG_FW_COMMAND(f...) IPW_DEBUG(IPW_DL_FW_COMMAND, ## f)
+#define IPW_DEBUG_HC(f...) IPW_DEBUG(IPW_DL_HC, ## f)
 #define IPW_DEBUG_FRAG(f...) IPW_DEBUG(IPW_DL_FRAG, ## f)
 #define IPW_DEBUG_FW(f...) IPW_DEBUG(IPW_DL_FW, ## f)
 #define IPW_DEBUG_RF_KILL(f...) IPW_DEBUG(IPW_DL_RF_KILL, ## f)
 #define IPW_DEBUG_DROP(f...) IPW_DEBUG(IPW_DL_DROP, ## f)
 #define IPW_DEBUG_IO(f...) IPW_DEBUG(IPW_DL_IO, ## f)
 #define IPW_DEBUG_IOCTL(f...) IPW_DEBUG(IPW_DL_IOCTL, ## f)
+#define IPW_DEBUG_STATE(f, a...) IPW_DEBUG(IPW_DL_STATE | IPW_DL_ASSOC | IPW_DL_INFO, f, ## a)
+#define IPW_DEBUG_ASSOC(f, a...) IPW_DEBUG(IPW_DL_ASSOC | IPW_DL_INFO, f, ## a)
 
 
 #define VERIFY(f) \
@@ -184,8 +177,8 @@ enum { IPW_DEBUG_ENABLED = 0 };
 }
 
 enum {
-	STATE_DISABLED = 1,
-	STATE_ENABLED = 0
+	IPW_HW_STATE_DISABLED = 1,
+	IPW_HW_STATE_ENABLED = 0
 };
 
 struct ssid_context {
@@ -200,10 +193,6 @@ struct ssid_context {
 extern const char *port_type_str[];
 extern const char *band_str[];
 
-#define DEFAULT_FRAG_THRESHOLD 2342U
-#define MIN_FRAG_THRESHOLD     256U
-#define	MAX_FRAG_THRESHOLD     2342U
-
 #define NUMBER_OF_BD_PER_COMMAND_PACKET		1
 #define NUMBER_OF_BD_PER_DATA_PACKET		2
 
@@ -322,9 +311,9 @@ struct ipw2100_wpa_assoc_frame {
 	u8 var_ie[IPW_MAX_VAR_IE_LEN];
 };
 
-#define BSS 1
-#define MONITOR 2
-#define IBSS 3
+#define IPW_BSS     1
+#define IPW_MONITOR 2
+#define IPW_IBSS    3
 
 /**
  * @struct _tx_cmd - HWCommand
@@ -419,7 +408,12 @@ struct ipw2100_rx_packet {
 	struct list_head list;
 };
 
-#define DEFAULT_RTS_THRESHOLD     2304U
+#define FRAG_DISABLED             BIT(31)
+#define RTS_DISABLED              BIT(31)
+#define MAX_RTS_THRESHOLD         2304U
+#define MIN_RTS_THRESHOLD         1U
+#define DEFAULT_RTS_THRESHOLD     1000U
+
 #define DEFAULT_BEACON_INTERVAL   100U
 #define	DEFAULT_SHORT_RETRY_LIMIT 7U
 #define	DEFAULT_LONG_RETRY_LIMIT  4U
@@ -445,12 +439,12 @@ struct ipw2100_notification {
 #define IPW2100_WEP_ENABLE     BIT(1)
 #define IPW2100_WEP_DROP_CLEAR BIT(2)
 
-#define IPW_NONE_CIPHER   BIT(0) 
-#define IPW_WEP40_CIPHER  BIT(1) 
-#define IPW_WEP104_CIPHER BIT(5) 
-#define IPW_TKIP_CIPHER   BIT(2) 
-#define IPW_CKIP_CIPHER   BIT(6)
+#define IPW_NONE_CIPHER   BIT(0)
+#define IPW_WEP40_CIPHER  BIT(1)
+#define IPW_TKIP_CIPHER   BIT(2)
 #define IPW_CCMP_CIPHER   BIT(4)
+#define IPW_WEP104_CIPHER BIT(5)
+#define IPW_CKIP_CIPHER   BIT(6)
 
 #define	IPW_AUTH_OPEN     0
 #define	IPW_AUTH_SHARED   1
@@ -488,80 +482,122 @@ enum {
 #endif
 };
 
-#define IPW_PREAMBLE_LONG 1
-#define IPW_PREAMBLE_AUTO 0
+#define STATUS_POWERED          BIT(0)
+#define STATUS_CMD_ACTIVE       BIT(1)  /**< host command in progress */
+#define STATUS_RUNNING          BIT(2)  /* Card initialized, but not enabled */
+#define STATUS_ENABLED          BIT(3)  /* Card enabled -- can scan,Tx,Rx */
+#define STATUS_STOPPING         BIT(4)  /* Card is in shutdown phase */
+#define STATUS_INITIALIZED      BIT(5)  /* Card is ready for external calls */
+#define STATUS_ASSOCIATING      BIT(9)  /* Associated, but no BSSID yet */
+#define STATUS_ASSOCIATED       BIT(10) /* Associated and BSSID valid */
+#define STATUS_INT_ENABLED      BIT(11)
+#define STATUS_RF_KILL_HW       BIT(12)
+#define STATUS_RF_KILL_SW       BIT(13)
+#define STATUS_RF_KILL_MASK     (STATUS_RF_KILL_HW | STATUS_RF_KILL_SW)
+#define STATUS_EXIT_PENDING     BIT(14)
+
+#define STATUS_SCAN_PENDING     BIT(23)
+#define STATUS_SCANNING         BIT(24) 
+#define STATUS_SCAN_ABORTING    BIT(25) 
+#define STATUS_SCAN_COMPLETE    BIT(26)
+#define STATUS_WX_EVENT_PENDING BIT(27)
+#define STATUS_RESET_PENDING    BIT(29)
+#define STATUS_SECURITY_UPDATED BIT(30) /* Security sync needed */
+
+
+
+/* Internal NIC states */
+#define IPW_STATE_INITIALIZED	BIT(0)
+#define IPW_STATE_COUNTRY_FOUND	BIT(1)
+#define IPW_STATE_ASSOCIATED    BIT(2)
+#define IPW_STATE_ASSN_LOST	BIT(3)
+#define IPW_STATE_ASSN_CHANGED 	BIT(4)
+#define IPW_STATE_SCAN_COMPLETE	BIT(5)
+#define IPW_STATE_ENTERED_PSP 	BIT(6)
+#define IPW_STATE_LEFT_PSP 	BIT(7)
+#define IPW_STATE_RF_KILL       BIT(8)
+#define IPW_STATE_DISABLED	BIT(9)
+#define IPW_STATE_POWER_DOWN	BIT(10)
+#define IPW_STATE_SCANNING      BIT(11)
+
+
+
+#define CFG_STATIC_CHANNEL      BIT(0) /* Restrict assoc. to single channel */
+#define CFG_STATIC_ESSID        BIT(1) /* Restrict assoc. to single SSID */
+#define CFG_STATIC_BSSID        BIT(2) /* Restrict assoc. to single BSSID */
+#define CFG_CUSTOM_MAC          BIT(3)
+#define CFG_LONG_PREAMBLE       BIT(4)
+#define CFG_ASSOCIATE           BIT(6)
+#define CFG_FIXED_RATE          BIT(7)
+#define CFG_ADHOC_CREATE        BIT(8)
+#define CFG_C3_DISABLED         BIT(9)
+#define CFG_PASSIVE_SCAN        BIT(10)
+
+#define CAP_SHARED_KEY          BIT(0) /* Off = OPEN */
+#define CAP_PRIVACY_ON          BIT(1) /* Off = No privacy */
 
 struct ipw2100_priv {
+
+	int stop_hang_check; /* Set 1 when shutting down to kill hang_check */
+	int stop_rf_kill; /* Set 1 when shutting down to kill rf_kill */
+	
 	struct ieee80211_device *ieee;
-	int reset_pending; /* 1 if a reset has been scheduled or is running */
+	unsigned long status;
+	unsigned long config;
+	unsigned long capability;
+
+	/* Statistics */
 	int resets;
 	int reset_backoff;
-	unsigned long last_reset;
-	int open;
-	int stopped;
-	int disabled;
-	int phy_off;
-	int connected;
-	int in_scan;
-	int manual_disable;
-	int preamble; /* 1 when long preamble is configured */
-	u32 channel_mask; 
-	int auto_associate;
-	int passive_scan;
+
+	/* Context */
+	u8 essid[IW_ESSID_MAX_SIZE];
+	u8 essid_len;
+	u8 bssid[ETH_ALEN];
+	u8 channel;
+	int last_mode;
+	int cstate_limit;
+
 	unsigned long connect_start;
-	int been_powered; /* 1 if HW has been powered */
+	unsigned long last_reset;
+	unsigned long start_rx;
+
+	u32 channel_mask; 
 	u32 fatal_error;
 	u32 fatal_errors[IPW2100_ERROR_QUEUE];
 	u32 fatal_index;
-	int rf_kill;
 	int eeprom_version;
 	int firmware_version;
 	unsigned long hw_features;
 	int hangs;
 	u32 last_rtc;
-	int dump_raw; /* 1 to dump raw bytes in /proc/.../memory */
+	int dump_raw; /* 1 to dump raw bytes in /sys/.../memory */
 	u8* snapshot[0x30];
 
-	int mandatory_bssid;
 	u8 mandatory_bssid_mac[ETH_ALEN];
-	int custom_addr;
 	u8 mac_addr[ETH_ALEN];
 	
 	int power_mode;
 
-	struct timer_list hang_check_timer;
-	int stop_hang_check; /* Set 1 when shutting down to kill timer */
-	struct timer_list rf_switch_timer;
-	int stop_rf_check; /* Set 1 when shutting down to kill timer */
- 
 	/* WEP data */
 	struct ieee80211_security sec;
-	int security_updated;
-	int wep_enabled;
-
-	int message_sending;
 	int messages_sent;
 	
 
 	int short_retry_limit;
 	int long_retry_limit;
 
-	int rts_threshold;
+	u32 rts_threshold;
 	u32 frag_threshold;
-	u8 frag_enabled;
+
 	int in_isr;
 
 	u32 tx_rates;
 	int tx_power;
 	u32 beacon_interval;
-	int port_type;
 
 	char nick[IW_ESSID_MAX_SIZE + 1];
 
-	struct ssid_context *ctx;
-	struct ssid_context active_ctx;
-	struct ssid_context backup_ctx;
-
 	struct ipw2100_status_queue status_queue;
 
 	struct statistic txq_stat;
@@ -587,45 +623,38 @@ struct ipw2100_priv {
 
 	struct ipw2100_ordinals ordinals;
 
-	struct pci_dev *pdev;
+	struct pci_dev *pci_dev;
 
 	struct proc_dir_entry *dir_dev;
 
-	struct net_device *ndev;
+	struct net_device *net_dev;
 	struct iw_statistics wstats;
 
 	struct tasklet_struct irq_tasklet;
-	int irq_ignore; /* set to 1 for IRQs to be ignored */
 
 	struct workqueue_struct *workqueue;
 	struct work_struct reset_work;
 	struct work_struct security_work;
 	struct work_struct wx_event_work;
+	struct work_struct hang_check;
+	struct work_struct rf_kill;
+	struct work_struct hw_kick;
 	
-	int wx_ap_event_pending;
-
-/*
-	struct sk_buff_head last_tx;
-	struct sk_buff_head last_rx;
-*/
-
 	u32 interrupts;
 	int tx_interrupts;
 	int rx_interrupts;
 	int inta_other;
 	
-	u32 proc_cardmemaddr;
-
 	spinlock_t low_lock;
+	struct semaphore action_sem;
+	struct semaphore adapter_sem;
 
 	wait_queue_head_t wait_command_queue;
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
 	u32 pm_state[PM_STATE_SIZE];
+#endif
 };
 
-static inline struct ipw2100_priv *ipw2100_netdev(struct net_device *dev) {
-	return (struct ipw2100_priv *)netdev_priv(dev);
-}
-
 
 /*********************************************************
  * Host Command -> From Driver to FW
@@ -685,24 +714,24 @@ static inline struct ipw2100_priv *ipw2100_netdev(struct net_device *dev) {
 
 
 // system configuration bit mask:
-//#define CFG_ANTENNA_SETTING           0x03
-//#define CFG_ANTENNA_A                 0x01
-//#define CFG_ANTENNA_B                 0x02
-#define CFG_PROMISCUOUS         0x00004
-//#define CFG_TX_STATUS_ENABLE    0x00008
-#define CFG_PREAMBLE_LEN        0x00010
-#define CFG_IBSS_AUTO_START     0x00020
-//#define CFG_KERBEROS_ENABLE     0x00040
-#define CFG_LOOPBACK            0x00100
-//#define CFG_WNMP_PING_PASS      0x00200
-//#define CFG_DEBUG_ENABLE        0x00400
-#define CFG_ANSWER_BCSSID_PROBE 0x00800
-//#define CFG_BT_PRIORITY         0x01000
-#define CFG_BT_SIDEBAND_SIGNAL	0x02000
-#define CFG_802_1x_ENABLE       0x04000
-#define CFG_BSS_MASK		0x08000
-#define CFG_IBSS_MASK		0x10000
-//#define CFG_DYNAMIC_CW          0x10000 
+//#define IPW_CFG_ANTENNA_SETTING           0x03
+//#define IPW_CFG_ANTENNA_A                 0x01
+//#define IPW_CFG_ANTENNA_B                 0x02
+#define IPW_CFG_PROMISCUOUS         0x00004
+//#define IPW_CFG_TX_STATUS_ENABLE    0x00008
+#define IPW_CFG_PREAMBLE_AUTO        0x00010
+#define IPW_CFG_IBSS_AUTO_START     0x00020
+//#define IPW_CFG_KERBEROS_ENABLE     0x00040
+#define IPW_CFG_LOOPBACK            0x00100
+//#define IPW_CFG_WNMP_PING_PASS      0x00200
+//#define IPW_CFG_DEBUG_ENABLE        0x00400
+#define IPW_CFG_ANSWER_BCSSID_PROBE 0x00800
+//#define IPW_CFG_BT_PRIORITY         0x01000
+#define IPW_CFG_BT_SIDEBAND_SIGNAL	0x02000
+#define IPW_CFG_802_1x_ENABLE       0x04000
+#define IPW_CFG_BSS_MASK		0x08000
+#define IPW_CFG_IBSS_MASK		0x10000
+//#define IPW_CFG_DYNAMIC_CW          0x10000 
 
 #define IPW_SCAN_NOASSOCIATE BIT(0)
 #define IPW_SCAN_MIXED_CELL BIT(1)
@@ -852,7 +881,6 @@ static inline struct ipw2100_priv *ipw2100_netdev(struct net_device *dev) {
 
 #define IPW_CACHE_LINE_LENGTH_DEFAULT		    0x80
 
-#define IPW_CARD_DISABLE_COMPLETE_WAIT		    100	// 100 milli
 #define IPW_CARD_DISABLE_PHY_OFF_COMPLETE_WAIT	    100	// 100 milli
 #define IPW_PREPARE_POWER_DOWN_COMPLETE_WAIT	    100	// 100 milli
 
@@ -908,9 +936,12 @@ struct ipw2100_rx {
 
 #define IPW_TX_POWER_AUTO            0
 #define IPW_TX_POWER_ENHANCED        1
+
 #define IPW_TX_POWER_DEFAULT         32
 #define IPW_TX_POWER_MIN             0
 #define IPW_TX_POWER_MAX             16
+#define IPW_TX_POWER_MIN_DBM         (-12)
+#define IPW_TX_POWER_MAX_DBM         16
 
 #define FW_SCAN_DONOT_ASSOCIATE     0x0001 // Dont Attempt to Associate after Scan
 #define FW_SCAN_PASSIVE             0x0008 // Force PASSSIVE Scan
@@ -933,19 +964,6 @@ struct ipw2100_rx {
 #define LOCK_DRIVER 1
 #define LOCK_FW 2
 
-#define STATUS_INIT_COMPLETE		0x0001
-#define STATUS_COUNTRY_FOUND		0x0002
-#define STATUS_ASSOCIATED		0x0004
-#define STATUS_ASSN_LOST		0x0008
-#define STATUS_ASSN_CHANGED 		0x0010
-#define STATUS_SCAN_COMPLETE		0x0020
-#define STATUS_ENTERED_PSP		0x0040
-#define STATUS_WAKEN_FROM_PSP		0x0080
-#define STATUS_RF_KILL_SWITCH		0x0100
-#define STATUS_CARD_DISABLE_COMPLETE	0x0200
-#define STATUS_PREP_PWR_DOWN_COMPLETE	0x0400
-#define STATUS_SCANNING 		0x0800	// STATUS_FW_SCAN_KEEP_ALIVE
-
 #define TYPE_SWEEP_ORD                  0x000D
 #define TYPE_IBSS_STTN_ORD              0x000E
 #define TYPE_BSS_AP_ORD                 0x000F
@@ -1207,245 +1225,6 @@ typedef enum _ORDINAL_TABLE_2 {	// NS - means Not Supported by FW
 
 #define IPW_LAST_VARIABLE_LENGTH_ORDINAL   1018
 
-extern int ipw2100_system_config(struct ipw2100_priv *, int);
-extern int ipw2100_get_ordinal(struct ipw2100_priv *, u32, void *, u32 *);
-extern int ipw2100_set_channel(struct ipw2100_priv *, u32, int);
-extern int ipw2100_set_mandatory_bssid(struct ipw2100_priv *, u8 *, int);
-extern int ipw2100_set_essid(struct ipw2100_priv *, char *, int, int);
-extern int ipw2100_set_power_mode(struct ipw2100_priv *, int);
-extern int ipw2100_set_rts_threshold(struct ipw2100_priv *, u32);
-extern int ipw2100_set_short_retry(struct ipw2100_priv *, u32);
-extern int ipw2100_set_long_retry(struct ipw2100_priv *, u32);
-extern int ipw2100_set_scan_options(struct ipw2100_priv *priv);
-extern int ipw2100_start_scan(struct ipw2100_priv *priv);
-#ifdef CONFIG_IPW2100_PROMISC
-extern int ipw2100_enable_monitor(struct ipw2100_priv *, u32);
-extern int ipw2100_disable_monitor(struct ipw2100_priv *);
-extern void ipw2100_reset_adapter(struct ipw2100_priv *);
-#endif
-extern int ipw2100_switch_mode(struct ipw2100_priv *, u32, u32);
-extern int ipw2100_set_tx_rates(struct ipw2100_priv *, u32, int);
-
-#include <linux/ctype.h>
-extern inline char *snprint_line(char *buf, size_t count, 
-				 u8 *data, u32 len, u32 ofs)
-{
-	int out, i, j, l;
-	char c;
-	
-	out = snprintf(buf, count, "%08X", ofs);
-
-	for (l = 0, i = 0; i < 2; i++) {
-		out += snprintf(buf + out, count - out, " ");
-		for (j = 0; j < 8 && l < len; j++, l++) 
-			out += snprintf(buf + out, count - out, "%02X ", 
-					data[(i * 8 + j)]);
-		for (; j < 8; j++)
-			out += snprintf(buf + out, count - out, "   ");
-	}
-	
-	out += snprintf(buf + out, count - out, " ");
-	for (l = 0, i = 0; i < 2; i++) {
-		out += snprintf(buf + out, count - out, " ");
-		for (j = 0; j < 8 && l < len; j++, l++) {
-			c = data[(i * 8 + j)];
-			if (!isascii(c) || !isprint(c))
-				c = '.';
-			
-			out += snprintf(buf + out, count - out, "%c", c);
-		}
-
-		for (; j < 8; j++)
-			out += snprintf(buf + out, count - out, " ");
-	}
-	
-	return buf;
-}
-
-extern inline void printk_buf(u8 *data, u32 len)
-{
-	char line[81];
-	u32 ofs = 0;
-
-	while (len) {
-		printk(KERN_DEBUG "%s\n",
-		       snprint_line(line, sizeof(line), &data[ofs], 
-				    min(len, 16U), ofs));
-		ofs += 16;
-		len -= min(len, 16U);
-	}
-}
-
-static inline void read_register(struct net_device *dev, u32 reg, u32 *val) 
-{
-	*val = readl((void *)(dev->base_addr + reg));
-	IPW_DEBUG_IO("r: 0x%08X => 0x%08X\n", reg, *val);
-}
-
-static inline void write_register(struct net_device *dev, u32 reg, u32 val) 
-{
-	writel(val, (void *)(dev->base_addr + reg));
-	IPW_DEBUG_IO("w: 0x%08X <= 0x%08X\n", reg, val);
-}
-
-static inline void read_register_word(struct net_device *dev, u32 reg, u16 *val)
-{
-	*val = readw((void *)(dev->base_addr + reg));
-	IPW_DEBUG_IO("r: 0x%08X => %04X\n", reg, *val);
-}
-
-static inline void read_register_byte(struct net_device *dev, u32 reg, u8 *val)
-{
-	*val = readb((void *)(dev->base_addr + reg));
-	IPW_DEBUG_IO("r: 0x%08X => %02X\n", reg, *val);
-}
-
-static inline void write_register_word(struct net_device *dev, u32 reg, u16 val)
-{
-	writew(val, (void *)(dev->base_addr + reg));
-	IPW_DEBUG_IO("w: 0x%08X <= %04X\n", reg, val);
-}
-
-
-static inline void write_register_byte(struct net_device *dev, u32 reg, u8 val)
-{
-	writeb(val, (void *)(dev->base_addr + reg));
-	IPW_DEBUG_IO("w: 0x%08X =< %02X\n", reg, val);
-}
-
-static inline void read_nic_dword(struct net_device *dev, u32 addr, u32 *val)
-{
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS,
-		       addr & IPW_REG_INDIRECT_ADDR_MASK);
-	read_register(dev, IPW_REG_INDIRECT_ACCESS_DATA, val);
-}
-
-static inline void write_nic_dword(struct net_device *dev, u32 addr, u32 val)
-{
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS,
-		       addr & IPW_REG_INDIRECT_ADDR_MASK);
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_DATA, val);
-}
-
-static inline void read_nic_word(struct net_device *dev, u32 addr, u16 *val)
-{
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS,
-		       addr & IPW_REG_INDIRECT_ADDR_MASK);
-	read_register_word(dev, IPW_REG_INDIRECT_ACCESS_DATA, val);
-}
-
-static inline void write_nic_word(struct net_device *dev, u32 addr, u16 val)
-{
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS,
-		       addr & IPW_REG_INDIRECT_ADDR_MASK);
-	write_register_word(dev, IPW_REG_INDIRECT_ACCESS_DATA, val);
-}
-
-static inline void read_nic_byte(struct net_device *dev, u32 addr, u8 *val)
-{
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS,
-		       addr & IPW_REG_INDIRECT_ADDR_MASK);
-	read_register_byte(dev, IPW_REG_INDIRECT_ACCESS_DATA, val);
-}
-
-static inline void write_nic_byte(struct net_device *dev, u32 addr, u8 val)
-{
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS,
-		       addr & IPW_REG_INDIRECT_ADDR_MASK);
-	write_register_byte(dev, IPW_REG_INDIRECT_ACCESS_DATA, val);
-}
-
-static inline void write_nic_auto_inc_address(struct net_device *dev, u32 addr)
-{
-	write_register(dev, IPW_REG_AUTOINCREMENT_ADDRESS,
-		       addr & IPW_REG_INDIRECT_ADDR_MASK);
-}
-
-static inline void write_nic_dword_auto_inc(struct net_device *dev, u32 val)
-{
-	write_register(dev, IPW_REG_AUTOINCREMENT_DATA, val);
-}
-
-static inline void write_nic_memory(struct net_device *dev, u32 addr, u32 len,
-				    const u8 *buf)
-{
-	u32 aligned_addr;
-	u32 aligned_len;
-	u32 dif_len;
-	u32 i;
-
-	/* read first nibble byte by byte */
-	aligned_addr = addr & (~0x3);
-	dif_len = addr - aligned_addr;
-	if (dif_len) {
-		/* Start reading at aligned_addr + dif_len */
-		write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS, 
-			       aligned_addr);
-		for (i = dif_len; i < 4; i++, buf++)
-			write_register_byte(
-				dev, IPW_REG_INDIRECT_ACCESS_DATA + i, 
-				*buf);
-		
-		len -= dif_len;
-		aligned_addr += 4;
-	}
-
-	/* read DWs through autoincrement registers */
-	write_register(dev, IPW_REG_AUTOINCREMENT_ADDRESS,
-		       aligned_addr);
-	aligned_len = len & (~0x3);
-	for (i = 0; i < aligned_len; i += 4, buf += 4, aligned_addr += 4)
-		write_register(
-			dev, IPW_REG_AUTOINCREMENT_DATA, *(u32 *)buf);
-	
-	/* copy the last nibble */
-	dif_len = len - aligned_len;
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS, aligned_addr);
-	for (i = 0; i < dif_len; i++, buf++)
-		write_register_byte(
-			dev, IPW_REG_INDIRECT_ACCESS_DATA + i, *buf);
-}
-
-static inline void read_nic_memory(struct net_device *dev, u32 addr, u32 len,
-				   u8 *buf)
-{
-	u32 aligned_addr;
-	u32 aligned_len;
-	u32 dif_len;
-	u32 i;
-
-	/* read first nibble byte by byte */
-	aligned_addr = addr & (~0x3);
-	dif_len = addr - aligned_addr;
-	if (dif_len) {
-		/* Start reading at aligned_addr + dif_len */
-		write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS, 
-			       aligned_addr);
-		for (i = dif_len; i < 4; i++, buf++)
-			read_register_byte(
-				dev, IPW_REG_INDIRECT_ACCESS_DATA + i, buf);
-		
-		len -= dif_len;
-		aligned_addr += 4;
-	}
-
-	/* read DWs through autoincrement registers */
-	write_register(dev, IPW_REG_AUTOINCREMENT_ADDRESS,
-		       aligned_addr);
-	aligned_len = len & (~0x3);
-	for (i = 0; i < aligned_len; i += 4, buf += 4, aligned_addr += 4)
-		read_register(dev, IPW_REG_AUTOINCREMENT_DATA,
-			      (u32 *)buf);
-	
-	/* copy the last nibble */
-	dif_len = len - aligned_len;
-	write_register(dev, IPW_REG_INDIRECT_ACCESS_ADDRESS,
-		       aligned_addr);
-	for (i = 0; i < dif_len; i++, buf++)
-		read_register_byte(dev, IPW_REG_INDIRECT_ACCESS_DATA +
-				   i, buf);
-}
-
 #ifndef WIRELESS_SPY
 #define WIRELESS_SPY		// enable iwspy support
 #endif
